 
## Citation
This code supports the paper:
Calafat, F. M., Frederikse, T., and Horsburgh, K. (2022). The Sources of Sea-Level Changes in the Mediterranean Sea since 1960, Journal of Geophysical Research Oceans. 

Please cite this paper when using this code.

# Overview

The code implements spatiotemporal Bayesian hierarchical modeling of sea-level observations from tide gauges and satellite altimetry in combination with sea-level fingerprints of
contemporary land-mass changes (GRD effects) and glacial isostatic adjustment (GIA). It makes estimates of sea-level changes as well as rates of change and quantifies the individual
contribution from sterodynamic changes, GRD effects, and GIA to sea-level changes at any arbitrary location (observed or unobserved). Inference in the model is performed using
Hamiltonian Monte Carlo as implemented by the Stan probabilistic programming language.
 
## Prerequisites

This code requires:
- CmdStanPy (https://mc-stan.org/cmdstanpy/)
- Python and various dependencies (Numpy and Scipy)

## How to use the code

The file **bhm_model.stan** contains the Stan code for the Bayesian Hierarchical model.
The file **master_bhm.py** shows how to load and prepare the input data, compile the Stan code, fit the hierarchical model, save the output to a file, and make predictions. Take this file as an example of how to use the code. 
Predictions on a finer grid can be made using the Python script (**bhm_prediction.py**), once the solutions from the Stan model are available. Running this script produces a large number of
outputs, which are saved to a file named "predictions_bhm.mat".

**Note**: You are welcome to contact me to report bugs

## License

Copyright (C) 2022 Francisco Mir Calafat.

Distributed under GPLv3 as described in LICENSE.txt.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.


