    
    functions{
        
        real euclidean_dist(real x1, real y1, real x2, real y2){
            real d;
            if(x1 == x2 && y1 == y2)
                d = 0;
            else
                d = sqrt((x2-x1)*(x2-x1) + (y2-y1)*(y2-y1))/10;
            return d;
        }
        
        matrix matern_52_mat(matrix r,matrix r_sq, real sig, real l){
            return square(sig)*(1 + (sqrt(5)/l)*r + (5/(3*square(l)))*r_sq) .* exp((-sqrt(5)/l)*r);
        }
        
         vector matern_52_vec(vector r,vector r_sq, real sig, real l){
            return square(sig)*(1 + (sqrt(5)/l)*r + (5/(3*square(l)))*r_sq) .* exp((-sqrt(5)/l)*r);
        } 
        
        real matern_52_scalar(real r,real r_sq, real sig, real l){
            return square(sig)*(1 + (sqrt(5)/l)*r + (5/(3*square(l)))*r_sq) * exp((-sqrt(5)/l)*r);
        }    
                
        vector nngp_chol_vec(matrix dx1,matrix dx2,matrix dx1_sq,matrix dx2_sq,real s,real l,int[,] indM,int M,int n,
                             vector a,real delta){
            matrix[M*M,n-1] C1;
            matrix[M,n-1] C2;
            vector[n] ac;
            real ssq = square(s);
            real Aii;
            real delta_sqrt = sqrt(delta);
            matrix[M,M] Ddelta = diag_matrix(rep_vector(delta,M));
            
            C1[1,1] = matern_52_scalar(dx1[1,1],dx1_sq[1,1],s,l);
            C2[1,1] = matern_52_scalar(dx2[1,1],dx2_sq[1,1],s,l);
            for(i in 2:(M-1)){
                C1[1:(i*i),i] = matern_52_vec(dx1[1:(i*i),i],dx1_sq[1:(i*i),i],s,l);
                C2[1:i,i] = matern_52_vec(dx2[1:i,i],dx2_sq[1:i,i],s,l);
            }
            C1[:,M:] = matern_52_mat(dx1[:,M:],dx1_sq[:,M:],s,l);
            C2[:,M:] = matern_52_mat(dx2[:,M:],dx2_sq[:,M:],s,l);
            
            ac[1] = a[1]*(s + delta_sqrt);
            Aii = -C2[1,1]/(C1[1,1] + delta);
            ac[2] = a[2]*sqrt(ssq + C2[1,1] * Aii) - Aii*ac[1];
            for(i in 3:n){
                int ie = (i-1)<M ? (i-1) : M;
                row_vector[ie] Ai;
                matrix[ie,ie] Li;
                Li = cholesky_decompose(to_matrix(C1[1:ie*ie,i-1],ie,ie) + Ddelta[1:ie,1:ie]);
                Ai = -mdivide_right_tri_low(mdivide_left_tri_low(Li, C2[1:ie,i-1])',Li);
                ac[i] = a[i]*sqrt(ssq + delta + Ai*C2[1:ie,i-1]) - Ai*ac[indM[i-1,1:ie]];
            }
            return ac;
        }
        
        matrix nngp_chol_mat(matrix dx1, matrix dx2,matrix dx1_sq,matrix dx2_sq, real s, real l, int[,] indM, int M, int n,
                             matrix a,real delta){
            matrix[M*M,n-1] C1;
            matrix[M,n-1] C2;
            matrix[rows(a),n] ac;
            real ssq = square(s);
            real Aii;
            real delta_sqrt = sqrt(delta);
            matrix[M,M] Ddelta = diag_matrix(rep_vector(delta,M));
            
            C1[1,1] = matern_52_scalar(dx1[1,1],dx1_sq[1,1],s,l);
            C2[1,1] = matern_52_scalar(dx2[1,1],dx2_sq[1,1],s,l);
            for(i in 2:(M-1)){
                C1[1:(i*i),i] = matern_52_vec(dx1[1:(i*i),i],dx1_sq[1:(i*i),i],s,l);
                C2[1:i,i] = matern_52_vec(dx2[1:i,i],dx2_sq[1:i,i],s,l);
            }
            C1[:,M:] = matern_52_mat(dx1[:,M:],dx1_sq[:,M:],s,l);
            C2[:,M:] = matern_52_mat(dx2[:,M:],dx2_sq[:,M:],s,l);
            
            ac[:,1] = a[:,1]*(s + delta_sqrt); 
            Aii = -C2[1,1]/(C1[1,1] + delta);
            ac[:,2] = a[:,2]*sqrt(ssq + C2[1,1] * Aii) - Aii*ac[:,1];
            for(i in 3:n){
                int ie = (i-1)<M ? (i-1) : M;
                row_vector[ie] Ai;
                matrix[ie,ie] Li;
                Li = cholesky_decompose(to_matrix(C1[1:ie*ie,i-1],ie,ie) + Ddelta[1:ie,1:ie]);
                Ai = -mdivide_right_tri_low(mdivide_left_tri_low(Li, C2[1:ie,i-1])',Li); 
                ac[:,i] = a[:,i]*sqrt(ssq + delta + Ai*C2[1:ie,i-1]) - ac[:,indM[i-1,1:ie]]*Ai'; 
            }
           
            return ac';
        }
    }
    
    data {
        int<lower=0> ntg; // number of TG data sites
        int<lower=0> na; // number of altimetry locations
        int<lower=0> nt; // length of TG timeseries
        int<lower=0> nta; // length of altimetry timeseries
        int<lower=0> nok;
        int<lower=0> noka;
        int<lower=0> M; // number of nearest neighbors for the GPs
        int<lower=0> K; // number of EOFs
        vector[ntg] lon_tg;
        vector[ntg] lat_tg;
        vector[na] lona;
        vector[na] lata;
        int kok[nok];
        int koka[noka];
        vector[nok] y_tg;
        vector[noka] y_alti;
        matrix[ntg,K] eof;
        matrix[na,K] eofa;
        matrix[ntg,nt] mur;
        matrix[na,nta] mug;
        vector[ntg] mu_giar;
        vector[na] mu_giag;
        matrix[ntg,ntg] Lr;
        matrix[na,na] Lg;
        matrix[ntg,ntg] Lr0;
        matrix[na,na] Lg0;
        matrix[ntg,ntg] Lgiar;
        matrix[na,na] Lgiag;
    }
    
    transformed data{
        int n = ntg + na;
        int nt0a = nt - nta + 1;
        vector[n] lon;
        vector[n] lat;
        matrix[M*M,n-1] dx1;
        matrix[M,n-1] dx2;
        matrix[M*M,n-1] dx1_sq;
        matrix[M,n-1] dx2_sq;
        int indM[n-1,M];
        matrix[M*M,ntg-1] dx1_;
        matrix[M,ntg-1] dx2_;
        matrix[M*M,ntg-1] dx1_sq_;
        matrix[M,ntg-1] dx2_sq_;
        int indM_[ntg-1,M];
        row_vector[nt] tv;
        matrix[n,nt] t;
        real delta = 1e-12;
        real dtmp;
        real tscale = 100;
        real s = 0.3;
        real s_uniform = 0.3;
        real l = 0.5;
        
        lon[1:ntg] = lon_tg;
        lon[(ntg+1):] = lona;
        lat[1:ntg] = lat_tg;
        lat[(ntg+1):] = lata;
        
        for(i in 1:nt)
            tv[i] = (i - 1)/tscale;
        tv = tv - mean(tv);
        t = rep_matrix(tv,n);
        
        for(i in 1:n-1){
            if(i<=M){
                for(j in 1:i){
                    indM[i,j] = j; 
                }
            }
            else{
                vector[i] dxi;
                for(j in 1:i)
                    dxi[j] = euclidean_dist(lon[j],lat[j],lon[i+1],lat[i+1]);
                indM[i,:] = sort_asc(sort_indices_asc(dxi)[1:M]);
            }
        }
        
        for(i in 1:n-1){
            if(i==1){
                dx1[1,i] = 0;
                dx2[1,i] = euclidean_dist(lon[1],lat[1],lon[2],lat[2]);
            }
            else{
                int ie = i<M ? i : M;
                int ct = 0;
                for(j in 1:ie){
                    for(k in 1:ie){
                        ct += 1;
                        dx1[ct,i] = euclidean_dist(lon[indM[i,j]],lat[indM[i,j]],lon[indM[i,k]],lat[indM[i,k]]);
                    }
                    dx2[j,i] = euclidean_dist(lon[indM[i,j]],lat[indM[i,j]],lon[i+1],lat[i+1]);
                }
            }   
        }
        dx1_sq = dx1 .* dx1;
        dx2_sq = dx2 .* dx2; 

        for(i in 1:ntg-1){
            if(i<=M){
                for(j in 1:i){
                    indM_[i,j] = j; 
                }
            }
            else{
                vector[i] dxi;
                for(j in 1:i)
                    dxi[j] = euclidean_dist(lon_tg[j],lat_tg[j],lon_tg[i+1],lat_tg[i+1]);
                indM_[i,:] = sort_asc(sort_indices_asc(dxi)[1:M]);
            }
        }
        
        for(i in 1:ntg-1){
            if(i==1){
                dx1_[1,i] = 0;
                dx2_[1,i] = euclidean_dist(lon_tg[1],lat_tg[1],lon_tg[2],lat_tg[2]);
            }
            else{
                int ie = i<M ? i : M;
                int ct = 0;
                for(j in 1:ie){
                    for(k in 1:ie){
                        ct += 1;
                        dx1_[ct,i] = euclidean_dist(lon_tg[indM_[i,j]],lat_tg[indM_[i,j]],lon_tg[indM_[i,k]],lat_tg[indM_[i,k]]);
                    }
                    dx2_[j,i] = euclidean_dist(lon_tg[indM_[i,j]],lat_tg[indM_[i,j]],lon_tg[i+1],lat_tg[i+1]);
                }
            }   
        }
        dx1_sq_ = dx1_ .* dx1_;
        dx2_sq_ = dx2_ .* dx2_; 
    }
    
    parameters {
        real<lower=0> s0;
        real<lower=0> l0;
        real<lower=0> s_small;
        vector[ntg] b_small;
        vector[na] b_small_a;
        real<lower=0> s_small_a;
        vector[n*nt] a;
        vector[n] a0;
        matrix[K,nt] pc;
        real<lower=0> s_pc;
        real<lower=0> sy_tg;
        real<lower=0> sy_alti;
        vector[n] interc;
        vector[ntg*(nt-1)] ar;
        vector[na*(nta-1)] ag;
        vector[ntg] ar0;
        vector[na] ag0;
        vector[ntg] a_giar;
        vector[na] a_giag;
        vector[nt] a_uniform;
        real b0_uniform;
    }
    
    transformed parameters {
        matrix[n,nt] b;
        matrix[n,nt] bi;
        matrix[ntg,nt] h_rel;
        matrix[na,nta] h_geo;
        vector[ntg] br0;
        vector[na] bg0;
        matrix[ntg,nt] br;
        matrix[na,nta] bg;
        matrix[ntg,nt-1] bri;
        matrix[na,nta-1] bgi;
        vector[ntg] b_giar;
        vector[na] b_giag;
        row_vector[nt] b_uniform;
        real su = s_uniform/1000; 
        
        b_uniform[nt] = b0_uniform/100 + su*a_uniform[nt];
        {
            int i = nt-1;
            while(i>=1){
                b_uniform[i] = b_uniform[i+1] + su*a_uniform[i];
                i -= 1;
            }
        }
        
        bi = nngp_chol_mat(dx1,dx2,dx1_sq,dx2_sq,s/1000,l,indM,M,n,to_matrix(a,nt,n,0),delta);
        b[:,nt] = nngp_chol_vec(dx1,dx2,dx1_sq,dx2_sq,s0/100,l0,indM,M,n,a0,delta) + bi[:,nt];
        
        br0 = Lr0 * ar0;
        bg0 = Lg0 * ag0;
        bri = Lr * to_matrix(ar,ntg,nt-1);
        bgi = Lg * to_matrix(ag,na,nta-1);
        b_giar = Lgiar * a_giar;
        b_giag = Lgiag * a_giag;
        br[:,1] = br0;
        bg[:,1] = bg0;

        for(i in 2:nt){
            br[:,i] = br[:,i-1] + bri[:,i-1];
            if(i<=nta)
                bg[:,i] = bg[:,i-1] + bgi[:,i-1];  
        }   
        
        {
            int i = nt-1;
            while(i>=1){
                b[:,i] = b[:,i+1] + bi[:,i];
                i-=1;
            }
        }
        
        h_rel[:,1] = interc[1:ntg];
        h_geo[:,1] = interc[(ntg+1):];
        for(i in 2:nt){
            h_rel[:,i] = h_rel[:,i-1] + (b_giar + mu_giar + br[:,i-1] + mur[:,i-1])/1000 + b[1:ntg,i-1] + b_uniform[i-1] + b_small;
            if(i<=nta)
                h_geo[:,i] = h_geo[:,i-1] + (b_giag + mu_giag + bg[:,i-1] + mug[:,i-1])/1000 + b[(ntg+1):,nt0a+i-2] + b_uniform[nt0a+i-2] + b_small_a;
        }    
    }
    
    model {
       s0 ~ normal(0,1);
       s_pc ~ normal(0,1);
       s_small ~ normal(0,1);
       b_small ~ normal(0,s_small/100);
       s_small_a ~ normal(0,1);
       b_small_a ~ normal(0,s_small_a/500);
       l0 ~ inv_gamma(5,2);
       a ~ std_normal();
       a0 ~ std_normal();
       to_vector(pc) ~ normal(0,s_pc*2);
       sy_tg ~ normal(0,1);
       sy_alti ~ normal(0,1);
       interc ~ normal(0,1);
       ar ~ std_normal();
       ag ~ std_normal();
       ar0 ~ std_normal();
       ag0 ~ std_normal();
       a_giar ~ std_normal();
       a_giag ~ std_normal();
       a_uniform ~ std_normal();
       b0_uniform ~ normal(0,1);

       y_tg ~ normal(to_vector(h_rel + eof*pc)[kok],sy_tg);
       y_alti ~ normal(to_vector(h_geo + eofa*pc[:,nt0a:])[koka],sy_alti);
    }
    
    generated quantities {
        real s_copy = s;
        real l_copy = l;
    }
   

