import numpy as np
from scipy.io import savemat,loadmat
from numpy.linalg import norm
from matern_52 import matern_52
from scipy.linalg import cholesky
from gp_pred import gp_pred
from gp_pred_land import gp_pred_land
from scipy.linalg import solve
from numpy.random import standard_normal
from multiprocessing.pool import ThreadPool
import gc
pool = ThreadPool(processes=10) # processes is the number of worker threads to use for multiprocessing

# this function is used for multiprocessing using ThreadPool
def fnngp(arg1,arg2,arg3,arg4,arg5):
    mi = arg1 @ solve(arg2,arg3)
    vi = np.square(arg4) - arg1 @ solve(arg2,arg1)
    return mi + np.sqrt(vi) * arg5

path2estimates = '/path/to/estimates/' # path to the Stan model output (path to the file generated by 'saveEstimates.py')
path2data = '/path2/to/data/' # path to data (observations and sea-level fingerprint)
pathOut = '/path/to/predictions/' # path where predictions will be saved
inFile = 'estimates_bhm.mat' # name of the file that contains the Stan model output (from 'saveEstimates.py')
outFile = 'predictions_bhm.mat' # name of the file where to save the predictions
data = loadmat(path2estimates + inFile)
data2 = loadmat(path2data + 'data_input.mat') # 'data_input.mat': file containing the observations and sea-level fingerprint data (available at https://doi.org/10.5281/zenodo.6797928)

nit = data["l_copy"].shape[0]
data["s"] = data["s_copy"]
data["l"] = data["l_copy"]
kok_pred_grid = data2["kok_pred_grid"][:,0] - 1
kok_est = data2["kok_est"][:,0] - 1
kok_pred = data2["kok_pred_noEst"][:,0] - 1 # indices of prediction sites corresponding to valid points in the full alti grid
loni = data2["lon_pred"][kok_pred,:] # lon_pred contains longitudes for the whole altimetry grid. kok_pred selects altimetry grid points with no estimates
lati = data2["lat_pred"][kok_pred,:]
nt_alti = data2["sla"].shape[2]
mug = data2["grd_mean_gsl"]
j1,j2,j3 = mug.shape
mug = np.reshape(mug,(j1*j2,j3),order='F')
kok_reduced = np.nonzero(np.mean(mug,1) == np.mean(mug,1))[0] # indices of valid points in the reduced alti grid
mug = mug[kok_reduced,:]
mug = mug[:,-nt_alti:]
delta = 1e-12 # used to deal with numerical non-positive definiteness

npr = loni.shape[0]
no,ntg = data["sea_level"].shape
na = data["sla"].shape[1]
nx = ntg + na
lon = np.zeros((nx,1))
lat = np.zeros((nx,1))
lon[0:ntg,0] = data["lon"][:,0]
lon[ntg:,0] = data["lona"][:,0]
lat[0:ntg,0] = data["lat"][:,0]
lat[ntg:,0] = data["lata"][:,0]

# compute distance
M = data["M"][0][0]
indM = np.zeros((M,npr),dtype=np.intp)
dx12 = np.zeros((M,npr))
dx = np.zeros((M,M,npr))
for i in range(npr):
    di = norm(np.array((lon[:,0],lat[:,0])) - np.array((loni[i,:],lati[i,:])),axis=0,ord=2)/10.
    inds = np.argsort(di)[0:M]
    dx12[:,i] = di[inds]
    indM[:,i] = inds
        
    lo = np.repeat(lon[inds,:],M,axis=1)
    la = np.repeat(lat[inds,:],M,axis=1)
    dx[:,:,i] = norm(np.array((lo,la)) - np.array((lo.T,la.T)),axis=0,ord=2)/10.
    lo = None
    la = None

# prediction for land-mass contribution (GRD and GIA)
giar_pred = data2["gia_mean_rsl_pred"].T + gp_pred_land(data["b_giar"],data2["gia_chol_rsl_tg"],data2["gia_cov_rsl_pred"],data2["gia_cov_rsl_tg_pred"],delta)
giar_pred = giar_pred/1000 # RSL

giag_predi = data2["gia_mean_gsl_pred"].T + gp_pred_land(data["b_giag"],data2["gia_chol_gsl"],data2["gia_cov_gsl_pred"],data2["gia_cov_gsl_est_pred"],delta)
giag_pred = np.zeros((nit,data2["lon_pred"].shape[0]))
giag_pred[:,kok_pred] = giag_predi/1000 # GSL
giag_pred[:,kok_est] = (data["b_giag"] + data2["gia_mean_gsl"].flatten('F')[kok_reduced])/1000

br_pred = np.zeros((nit,data2["lon_pred"].shape[0],no)) # RSL
br_pred[:,:,0] = gp_pred_land(data["br0"],data2["grd_chol_rsl_init_tg"],data2["grd_cov_rsl_init_pred"],data2["grd_cov_rsl_init_tg_pred"],delta)
bg_pred = np.zeros((nit,data2["lon_pred"].shape[0],nt_alti)) # GSL
bg_pred[:,kok_pred,0] = gp_pred_land(data["bg0"],data2["grd_chol_gsl_init"],data2["grd_cov_gsl_init_pred"],data2["grd_cov_gsl_init_est_pred"],delta)
for i in range(1,no):
    br_pred[:,:,i] = br_pred[:,:,i-1] + gp_pred_land(data["bri"][:,:,i-1],data2["grd_chol_rsl_tg"],data2["grd_cov_rsl_pred"],data2["grd_cov_rsl_tg_pred"],delta)
    if i<nt_alti:
        bg_pred[:,kok_pred,i] = bg_pred[:,kok_pred,i-1] + gp_pred_land(data["bgi"][:,:,i-1],data2["grd_chol_gsl"],data2["grd_cov_gsl_pred"],data2["grd_cov_gsl_est_pred"],delta)
br_pred = np.reshape(br_pred,(nit,data2["lon_pred"].shape[0]*no),order='F')
br_pred = (br_pred + (data2["grd_mean_rsl_pred"].flatten('F')).T)/1000
br_pred = np.reshape(br_pred,(nit,data2["lon_pred"].shape[0],no),order='F')

bg_predi = np.reshape(bg_pred[:,kok_pred,:],(nit,kok_pred.shape[0]*nt_alti),order='F')
bg_predi = (bg_predi + (data2["grd_mean_gsl_pred"][:,-nt_alti:].flatten('F')).T)/1000
bg_pred[:,kok_pred,:] = np.reshape(bg_predi,(nit,kok_pred.shape[0],nt_alti),order='F')
bg_predi = (np.reshape(data["bg"],(nit,na*nt_alti),order='F') + (mug.flatten('F')).T)/1000
bg_pred[:,kok_est,:] = np.reshape(bg_predi,(nit,na,nt_alti),order='F')

# prediction for sterodynamic
b_pred = np.zeros((nit,npr,no))
bi_pred = np.zeros((nit,npr,no))
b0_pred = np.zeros((nit,npr))
Ddelta = np.identity(M)*delta
Q_b0p = matern_52(dx12,data["s0"]/100,data["l0"])
Q_bp = matern_52(dx12,data["s"]/1000,data["l"])
rn0 = standard_normal((nit,npr))
rn = standard_normal((nit,npr,no))
for i in range(npr):
    if np.mod(i,50) == 0:
        print(i)
    Q_b0 = matern_52(dx[:,:,i],data["s0"]/100,data["l0"]) + Ddelta
    Q_b = matern_52(dx[:,:,i],data["s"]/1000,data["l"]) + Ddelta
    arg1 = (data["b"][:,:,-1] - data["bi"][:,:,-1])[:,indM[:,i]]
    bic = data["bi"][:,indM[:,i],:]
        
    # we use multiprocessing
    def fparal0(j):
        return fnngp(Q_b0p[j,:,i],Q_b0[j,:,:],arg1[j,:],data["s0"][j]/100,rn0[j,i])
            
    def fparali(j):
        return fnngp(Q_bp[j,:,i],Q_b[j,:,:],bic[j,:,:],data["s"][j]/1000,rn[j,i,:])
            
    b0_pred[:,i] = pool.map(fparal0, range(nit)) # multiprocessing
    bi_pred[:,i,:] = pool.map(fparali, range(nit))
            
b_pred[:,:,-1] = b0_pred + bi_pred[:,:,-1]  
for j in range(no-2,-1,-1):
    b_pred[:,:,j] = b_pred[:,:,j+1] + bi_pred[:,:,j]

bi_pred = None
Q_bp = None
Q_b0p = None


n1 = int(data2["size_grid"][0][0])
n2 = int(data2["size_grid"][0][1])
bb = np.zeros((nit,data2["lon_pred"].shape[0],no)) # sterodynamic full grid
bb[:,data2["kok_est"][:,0]-1,:] = data["b"][:,ntg:,:]
bb[:,data2["kok_pred_noEst"][:,0]-1,:] = b_pred
br_grid = np.zeros((nit,n1*n2,no))*np.NaN
br_grid[:,kok_pred_grid,:] = br_pred
br_grid = np.reshape(br_grid,(nit,n1*n2,no),order='F')
bg_grid = np.zeros((nit,n1*n2,nt_alti))*np.NaN
bg_grid[:,kok_pred_grid,:] = bg_pred
bg_grid = np.reshape(bg_grid,(nit,n1*n2,nt_alti),order='F')
bgiar_grid = np.zeros((nit,n1*n2))*np.NaN
bgiar_grid[:,kok_pred_grid] = giar_pred
bgiar_grid = np.reshape(bgiar_grid,(nit,n1*n2),order='F')
bgiag_grid = np.zeros((nit,n1*n2))*np.NaN
bgiag_grid[:,kok_pred_grid] = giag_pred
bgiag_grid = np.reshape(bgiag_grid,(nit,n1*n2),order='F')
h_rel = np.zeros((nit,data2["lon_pred"].shape[0],no))
h_gsl = np.zeros((nit,data2["lon_pred"].shape[0],nt_alti)) # GSL
h_dyn = np.zeros((nit,data2["lon_pred"].shape[0],no))
h_grd = np.zeros((nit,data2["lon_pred"].shape[0],no))
h_grdg = np.zeros((nit,data2["lon_pred"].shape[0],nt_alti)) # GRD GSL
h_uni = np.zeros((nit,no))
for i in range(1,no):
    h_rel[:,:,i] = h_rel[:,:,i-1] + giar_pred + br_pred[:,:,i-1] + bb[:,:,i-1] + data["b_uniform"][:,i-1][:,None]
    h_dyn[:,:,i] = h_dyn[:,:,i-1] + bb[:,:,i-1] + data["b_uniform"][:,i-1][:,None]
    h_grd[:,:,i] = h_grd[:,:,i-1] + br_pred[:,:,i-1]
    h_uni[:,i] = h_uni[:,i-1] + data["b_uniform"][:,i-1]
    if i<nt_alti:
        ii = no - nt_alti + i - 1
        h_gsl[:,:,i] = h_gsl[:,:,i-1] + giag_pred + bg_pred[:,:,i-1] + bb[:,:,ii] + data["b_uniform"][:,ii][:,None]
        h_grdg[:,:,i] = h_grdg[:,:,i-1] + bg_pred[:,:,i-1]

h_dyn_rate = np.zeros((nit,data2["lon_pred"].shape[0],no))
h_total_rate = np.zeros((nit,data2["lon_pred"].shape[0],no))
h_total_rateg = np.zeros((nit,data2["lon_pred"].shape[0],nt_alti))
for i in range(no):
    h_dyn_rate[:,:,i] = bb[:,:,i] + data["b_uniform"][:,i][:,None] 
    h_total_rate[:,:,i] = giar_pred + br_pred[:,:,i] + bb[:,:,i] + data["b_uniform"][:,i][:,None]
    if i<nt_alti:
        ii = no - nt_alti + i
        h_total_rateg[:,:,i] = giag_pred + bg_pred[:,:,i] + bb[:,:,ii] + data["b_uniform"][:,ii][:,None]
    
bb = None
h_dyn_rate_grid = np.zeros((nit,n1*n2,no))*np.NaN
h_dyn_rate_grid[:,kok_pred_grid,:] = h_dyn_rate
h_dyn_rate_basin = np.mean(np.mean(h_dyn_rate,1),0)
h_dyn_rate_basin_sd = np.std(np.mean(h_dyn_rate,1),0)
h_dyn_rate_basin_samples = np.mean(h_dyn_rate,1)

# regional analysis
kreg1 = data2["lon_pred"][:,0] >=29
h_dyn_rate_levantine = np.mean(h_dyn_rate[:,kreg1,:],1)
kreg2 = (data2["lon_pred"][:,0] >=22.5) & (data2["lon_pred"][:,0] <=28) & (data2["lat_pred"][:,0] >=35.5) & (data2["lat_pred"][:,0] <=41.5)
kreg3 = (data2["lon_pred"][:,0] <10) | (data2["lon_pred"][:,0] >=10) & (data2["lon_pred"][:,0] <12.3) & (data2["lat_pred"][:,0] >37.3) & (data2["lat_pred"][:,0] <45) |\
        (data2["lon_pred"][:,0] >=12.3) & (data2["lon_pred"][:,0] <14) & (data2["lat_pred"][:,0] >38) & (data2["lat_pred"][:,0] <42.5) |\
        (data2["lon_pred"][:,0] >=14) & (data2["lon_pred"][:,0] <16.2) & (data2["lat_pred"][:,0] >38) & (data2["lat_pred"][:,0] <41.2)
kreg4 = (data2["lon_pred"][:,0] >=12) & (data2["lon_pred"][:,0] <19) & (data2["lat_pred"][:,0] >42.5) & (data2["lat_pred"][:,0] <46.5) |\
        (data2["lon_pred"][:,0] >14) & (data2["lon_pred"][:,0] <20) & (data2["lat_pred"][:,0] >41.7) & (data2["lat_pred"][:,0] <=42.5) |\
        (data2["lon_pred"][:,0] >15.7) & (data2["lon_pred"][:,0] <20) & (data2["lat_pred"][:,0] >40.3) & (data2["lat_pred"][:,0] <=41.7) 
kreg5 = ~kreg1 & ~kreg2 & ~kreg3 & ~kreg4
h_dyn_rate_aegean = np.mean(h_dyn_rate[:,kreg2,:],1)
h_dyn_rate_western = np.mean(h_dyn_rate[:,kreg3,:],1)
h_dyn_rate_adri = np.mean(h_dyn_rate[:,kreg4,:],1)
h_dyn_rate_central = np.mean(h_dyn_rate[:,kreg5,:],1)
h_grd_rate_levantine = np.mean(br_pred[:,kreg1,:],1)
h_grd_rate_aegean = np.mean(br_pred[:,kreg2,:],1)
h_grd_rate_western = np.mean(br_pred[:,kreg3,:],1)
h_grd_rate_adri = np.mean(br_pred[:,kreg4,:],1)
h_grd_rate_central = np.mean(br_pred[:,kreg5,:],1)
h_gia_rate_levantine = np.mean(giar_pred[:,kreg1],1)
h_gia_rate_aegean = np.mean(giar_pred[:,kreg2],1)
h_gia_rate_western = np.mean(giar_pred[:,kreg3],1)
h_gia_rate_adri = np.mean(giar_pred[:,kreg4],1)
h_gia_rate_central = np.mean(giar_pred[:,kreg5],1)
kreg = kreg1 | kreg2
h_dyn_rate_levantineANDaegean = np.mean(h_dyn_rate[:,kreg,:],1)
q_tmp = h_dyn_rate.copy()
q2_tmp = np.mean(h_dyn_rate,1)
for i in range(no):
    q_tmp[:,:,i] = q_tmp[:,:,i] - q2_tmp[:,i][:,None]
h_dyn_rate_levantine_dev = np.mean(q_tmp[:,kreg1,:],1)
h_dyn_rate_aegean_dev = np.mean(q_tmp[:,kreg2,:],1)
h_dyn_rate_western_dev = np.mean(q_tmp[:,kreg3,:],1)
h_dyn_rate_adri_dev = np.mean(q_tmp[:,kreg4,:],1)
h_dyn_rate_central_dev = np.mean(q_tmp[:,kreg5,:],1)
h_dyn_rate_levantineANDaegean_dev = np.mean(q_tmp[:,kreg,:],1)
q_tmp = None
q2_tmp = None
gc.collect()
#--------------------
h_dyn_rate = None
h_dyn_grid = np.zeros((nit,n1*n2,no))*np.NaN
h_dyn_grid[:,kok_pred_grid,:] = h_dyn 
h_dyn_basin = np.mean(np.mean(h_dyn,1),0)
h_dyn_basin_sd = np.std(np.mean(h_dyn,1),0)
h_dyn = None
h_grd_grid = np.zeros((nit,n1*n2,no))*np.NaN
h_grd_grid[:,kok_pred_grid,:] = h_grd
h_grd_basin = np.mean(np.mean(h_grd,1),0)
h_grd_basin_sd = np.std(np.mean(h_grd,1),0)
h_grd = None
h_grdg_grid = np.zeros((nit,n1*n2,nt_alti))*np.NaN
h_grdg_grid[:,kok_pred_grid,:] = h_grdg
h_grdg_basin = np.mean(np.mean(h_grdg,1),0)
h_grdg_basin_sd = np.std(np.mean(h_grdg,1),0)
h_grdg = None
h_total_rate_grid = np.zeros((nit,n1*n2,no))*np.NaN
h_total_rate_grid[:,kok_pred_grid,:] = h_total_rate
h_total_rate_basin = np.mean(np.nanmean(h_total_rate,1),0)
h_total_rate_basin_sd = np.std(np.nanmean(h_total_rate,1),0)
h_total_rate_basin_samples = np.nanmean(h_total_rate,1)
# regional analysis
h_total_rate_levantine = np.mean(h_total_rate[:,kreg1,:],1)
h_total_rate_aegean = np.mean(h_total_rate[:,kreg2,:],1)
h_total_rate_western = np.mean(h_total_rate[:,kreg3,:],1)
h_total_rate_adri = np.mean(h_total_rate[:,kreg4,:],1)
h_total_rate_central = np.mean(h_total_rate[:,kreg5,:],1)
h_total_rate_levantineANDaegean = np.mean(h_total_rate[:,kreg,:],1)
q_tmp = h_total_rate.copy()
q2_tmp = np.mean(h_total_rate,1)
for i in range(no):
    q_tmp[:,:,i] = q_tmp[:,:,i] - q2_tmp[:,i][:,None]
h_total_rate_levantine_dev = np.mean(q_tmp[:,kreg1,:],1)
h_total_rate_aegean_dev = np.mean(q_tmp[:,kreg2,:],1)
h_total_rate_western_dev = np.mean(q_tmp[:,kreg3,:],1)
h_total_rate_adri_dev = np.mean(q_tmp[:,kreg4,:],1)
h_total_rate_central_dev = np.mean(q_tmp[:,kreg5,:],1)
h_total_rate_levantineANDaegean_dev = np.mean(q_tmp[:,kreg,:],1)
q_tmp = None
q2_tmp = None
gc.collect()
#--------------------
h_total_rate = None
h_total_rateg_grid = np.zeros((nit,n1*n2,nt_alti))*np.NaN
h_total_rateg_grid[:,kok_pred_grid,:] = h_total_rateg
h_total_rateg_basin = np.mean(np.nanmean(h_total_rateg,1),0)
h_total_rateg_basin_sd = np.std(np.nanmean(h_total_rateg,1),0)
h_total_rateg_basin_samples = np.nanmean(h_total_rateg,1)
h_total_rateg = None
gc.collect()

h_dyn_rate_grid = np.reshape(h_dyn_rate_grid,(nit,n1,n2,no),order='F') 
h_dyn_grid = np.reshape(h_dyn_grid,(nit,n1,n2,no),order='F')
h_grd_grid = np.reshape(h_grd_grid,(nit,n1,n2,no),order='F')
h_total_rate_grid = np.reshape(h_total_rate_grid,(nit,n1,n2,no),order='F')
h_grdg_grid = np.reshape(h_grdg_grid,(nit,n1,n2,nt_alti),order='F')
h_total_rateg_grid = np.reshape(h_total_rateg_grid,(nit,n1,n2,nt_alti),order='F')

#start storing data ****************************************
br_grid = np.reshape(br_grid,(nit,n1,n2,no),order='F')
bg_grid = np.reshape(bg_grid,(nit,n1,n2,nt_alti),order='F')
data4 = {}
data4["sterodyn_rate_rsl"] = np.mean(h_dyn_rate_grid,0)
data4["grd_rate_rsl"] = np.mean(br_grid,0)
data4["grd_rate_gsl"] = np.mean(bg_grid,0)
data4["sterodyn_rate_rsl_sd"] = np.std(h_dyn_rate_grid,0)
data4["grd_rate_rsl_sd"] = np.std(br_grid,0)
data4["grd_rate_gsl_sd"] = np.std(bg_grid,0)
data4["sterodyn_rsl"] = np.mean(h_dyn_grid,0)
data4["grd_rsl"] = np.mean(h_grd_grid,0)
data4["grd_gsl"] = np.mean(h_grdg_grid,0)
data4["sterodyn_rsl_sd"] = np.std(h_dyn_grid,0)
data4["grd_rsl_sd"] = np.std(h_grd_grid,0)
data4["grd_gsl_sd"] = np.std(h_grdg_grid,0)
data4["rsl_rate"] = np.mean(h_total_rate_grid,0)
data4["rsl_rate_sd"] = np.std(h_total_rate_grid,0)
data4["gsl_rate"] = np.mean(h_total_rateg_grid,0)
data4["gsl_rate_sd"] = np.std(h_total_rateg_grid,0)
data4["sterodyn_rate_rsl_levantine_samples"] = h_dyn_rate_levantine
data4["sterodyn_rate_rsl_aegean_samples"] = h_dyn_rate_aegean
data4["sterodyn_rate_rsl_western_samples"] = h_dyn_rate_western
data4["sterodyn_rate_rsl_adri_samples"] = h_dyn_rate_adri
data4["sterodyn_rate_rsl_central_samples"] = h_dyn_rate_central
data4["grd_rate_rsl_levantine_samples"] = h_grd_rate_levantine
data4["grd_rate_rsl_aegean_samples"] = h_grd_rate_aegean
data4["grd_rate_rsl_western_samples"] = h_grd_rate_western
data4["grd_rate_rsl_adri_samples"] = h_grd_rate_adri
data4["grd_rate_rsl_central_samples"] = h_grd_rate_central
data4["gia_rate_rsl_levantine_samples"] = h_gia_rate_levantine
data4["gia_rate_rsl_aegean_samples"] = h_gia_rate_aegean
data4["gia_rate_rsl_western_samples"] = h_gia_rate_western
data4["gia_rate_rsl_adri_samples"] = h_gia_rate_adri
data4["gia_rate_rsl_central_samples"] = h_gia_rate_central
data4["sterodyn_rate_rsl_levantineAndAegean_samples"] = h_dyn_rate_levantineANDaegean
data4["rsl_rate_levantine_samples"] = h_total_rate_levantine
data4["rsl_rate_aegean_samples"] = h_total_rate_aegean
data4["rsl_rate_western_samples"] = h_total_rate_western
data4["rsl_rate_adri_samples"] = h_total_rate_adri
data4["rsl_rate_central_samples"] = h_total_rate_central
data4["rsl_rate_levantineAndAegean_samples"] = h_total_rate_levantineANDaegean
data4["sterodyn_rate_rsl_levantine_dev_samples"] = h_dyn_rate_levantine_dev
data4["sterodyn_rate_rsl_aegean_dev_samples"] = h_dyn_rate_aegean_dev
data4["sterodyn_rate_rsl_western_dev_samples"] = h_dyn_rate_western_dev
data4["sterodyn_rate_rsl_adri_dev_samples"] = h_dyn_rate_adri_dev
data4["sterodyn_rate_rsl_central_dev_samples"] = h_dyn_rate_central_dev
data4["sterodyn_rate_rsl_levantineAndAegean_dev_samples"] = h_dyn_rate_levantineANDaegean_dev
data4["rsl_rate_levantine_dev_samples"] = h_total_rate_levantine_dev
data4["rsl_rate_aegean_dev_samples"] = h_total_rate_aegean_dev
data4["rsl_rate_western_dev_samples"] = h_total_rate_western_dev
data4["rsl_rate_adri_dev_samples"] = h_total_rate_adri_dev
data4["rsl_rate_central_dev_samples"] = h_total_rate_central_dev
data4["rsl_rate_levantineAndAegean_dev_samples"] = h_total_rate_levantineANDaegean_dev

br_grid = np.reshape(br_grid,(nit,n1*n2,no),order='F')
bg_grid = np.reshape(bg_grid,(nit,n1*n2,nt_alti),order='F')
h_grd_rate_basin = np.mean(np.nanmean(br_grid,1),0)
h_grd_rate_basin_sd = np.std(np.nanmean(br_grid,1),0)
h_grd_rate_basin_samples = np.nanmean(br_grid,1)
h_grd_rateg_basin = np.mean(np.nanmean(bg_grid,1),0)
h_grd_rateg_basin_sd = np.std(np.nanmean(bg_grid,1),0)
h_grd_rateg_basin_samples = np.nanmean(bg_grid,1)
data4["basin_mean_sterodyn_rate"] = h_dyn_rate_basin
data4["basin_mean_sterodyn_rate_samples"] = h_dyn_rate_basin_samples
data4["basin_mean_sterodyn_rate_sd"] = h_dyn_rate_basin_sd
data4["basin_mean_sterodyn_rsl"] = h_dyn_basin
data4["basin_mean_sterodyn_rsl_sd"] = h_dyn_basin_sd
data4["basin_mean_grd_rate"] = h_grd_rate_basin
data4["basin_mean_grd_rate_samples"] = h_grd_rate_basin_samples
data4["basin_mean_grd_rate_sd"] = h_grd_rate_basin_sd
data4["basin_mean_grd_rsl"] = h_grd_basin
data4["basin_mean_grd_rsl_sd"] = h_grd_basin_sd
data4["basin_mean_grd_rate_gsl"] = h_grd_rateg_basin
data4["basin_mean_grd_rate_gsl_samples"] = h_grd_rateg_basin_samples
data4["basin_mean_grd_rate_gsl_sd"] = h_grd_rateg_basin_sd
data4["basin_mean_grd_gsl"] = h_grdg_basin
data4["basin_mean_grd_gsl_sd"] = h_grdg_basin_sd
data4["basin_mean_rsl_rate"] = h_total_rate_basin
data4["basin_mean_rsl_rate_samples"] = h_total_rate_basin_samples
data4["basin_mean_rsl_rate_sd"] = h_total_rate_basin_sd
data4["basin_mean_gsl_rate"] = h_total_rateg_basin
data4["basin_mean_gsl_rate_samples"] = h_total_rateg_basin_samples
data4["basin_mean_gsl_rate_sd"] = h_total_rateg_basin_sd
data4["uniform_rsl"] = np.mean(h_uni,0)
data4["uniform_rsl_sd"] = np.std(h_uni,0)

h_dyn_rate_grid = None
h_grd_grid = None
h_grdg_grid = None
br_grid = None
bg_grid = None
h_total_rate_grid = None
h_total_rateg_grid = None
gc.collect()
#***********************************************

rsl = np.zeros((nit,n1*n2,no))*np.NaN
rsl[:,kok_pred_grid,:] = h_rel
gsl = np.zeros((nit,n1*n2,nt_alti))*np.NaN
gsl[:,kok_pred_grid,:] = h_gsl
h_rsl_basin = np.mean(np.mean(h_rel,1),0)
h_rsl_basin_sd = np.std(np.mean(h_rel,1),0)
h_rel = None
h_gsl_basin = np.mean(np.mean(h_gsl,1),0)
h_gsl_basin_sd = np.std(np.mean(h_gsl,1),0)
h_gsl = None

K = data["K"][0][0]
vari = np.zeros((nit,n1*n2,no))*np.NaN
for i in range(nit):
    vari[i,kok_pred_grid,:] = data2["eof_pred"][:,0:K] @ data["pc"][i,:,:]
vari = np.reshape(vari,(nit,n1,n2,no),order='F')
bgiar_grid = np.reshape(bgiar_grid,(nit,n1,n2),order='F')
rsl = np.reshape(rsl,(nit,n1,n2,no),order='F')
bgiag_grid = np.reshape(bgiag_grid,(nit,n1,n2),order='F')
gsl = np.reshape(gsl,(nit,n1,n2,nt_alti),order='F')

# store data *******************************************
data4["rsl"] = np.mean(rsl,0)
data4["rsl_sd"] = np.std(rsl,0)
data4["gsl"] = np.mean(gsl,0)
data4["gsl_sd"] = np.std(gsl,0)
#**********************************************************

vari2 = np.reshape(vari,(nit,n1*n2,no),order='F') 
rsl = np.reshape(rsl,(nit,n1*n2,no),order='F')
gsl = np.reshape(gsl,(nit,n1*n2,nt_alti),order='F')
total_rsl = rsl + vari2
total_gsl = gsl + vari2[:,:,-nt_alti:]
gc.collect()
# store data ********
data4["basin_mean_rsl"] = np.mean(np.nanmean(rsl,1),0)
data4["basin_mean_rsl_sd"] = np.std(np.nanmean(rsl,1),0)
data4["basin_mean_gsl"] = np.mean(np.nanmean(gsl,1),0)
data4["basin_mean_gsl_sd"] = np.std(np.nanmean(gsl,1),0)

data4["basin_mean_total_rsl"] = np.mean(np.nanmean(total_rsl,1),0)
data4["basin_mean_total_rsl_sd"] = np.std(np.nanmean(total_rsl,1),0)
data4["basin_mean_total_gsl"] = np.mean(np.nanmean(total_gsl,1),0)
data4["basin_mean_total_gsl_sd"] = np.std(np.nanmean(total_gsl,1),0)

total_rsl = np.reshape(total_rsl,(nit,n1,n2,no),order='F')
data4["total_rsl"] = np.mean(total_rsl,0)
data4["total_rsl_sd"] = np.std(total_rsl,0)
total_gsl = np.reshape(total_gsl,(nit,n1,n2,nt_alti),order='F')
data4["total_gsl"] = np.mean(total_gsl,0)
data4["total_gsl_sd"] = np.std(total_gsl,0)
rsl = None
total_rsl = None
gsl = None
total_gsl = None
gc.collect()

# ********

data4["gia_rate_rsl"] = np.mean(bgiar_grid,0)
data4["gia_rate_gsl"] = np.mean(bgiag_grid,0)
bgiar_grid = np.reshape(bgiar_grid,(nit,n1*n2),order='F')
bgiag_grid = np.reshape(bgiag_grid,(nit,n1*n2),order='F')
data4["basin_mean_gia_rsl"] = np.mean(np.nanmean(bgiar_grid,1),0)
data4["basin_mean_gia_rsl_samples"] = np.nanmean(bgiar_grid,1)
data4["basin_mean_gia_rsl_sd"] = np.std(np.nanmean(bgiar_grid,1),0)
data4["basin_mean_gia_gsl"] = np.mean(np.nanmean(bgiag_grid,1),0)
data4["basin_mean_gia_gsl_samples"] = np.nanmean(bgiag_grid,1)
data4["basin_mean_gia_gsl_sd"] = np.std(np.nanmean(bgiag_grid,1),0)
data4["uniform_rate_rsl"] = np.mean(data["b_uniform"],0)
data4["rsl_variability"] = np.mean(vari,0)
data4["rsl_variability_sd"] = np.std(vari,0)
data4["gia_rate_rsl_sd"] = np.std(bgiar_grid,0)
data4["gia_rate_gsl_sd"] = np.std(bgiag_grid,0)
data4["uniform_rate_rsl_sd"] = np.std(data["b_uniform"],0)
data4["basin_mean_variability_rsl"] = np.mean(np.nanmean(vari2,1),0)
data4["basin_mean_variability_rsl_sd"] = np.std(np.nanmean(vari2,1),0)
data4["lon"] = data2["lon_grid"]
data4["lat"] = data2["lat_grid"]
savemat(pathOut + outFile,data4,oned_as='column')
        
